"""
Config file that reads from .env file and parses based on type set.

https://docs.pydantic.dev/usage/settings/
"""
from urllib.parse import urlparse

import pytz
import requests
from pydantic import AnyUrl, BaseSettings, Field, validate_model, validator


class BaseConfig(BaseSettings):
    """
    Settings for project.
    """

    COPYFACTORY_API_KEY: str = Field(
        description="The Copyfactory global API key.", default=None
    )
    SALES_SCRAPERS_API_KEY: str = Field(
        description="The SalesScrapers API key", default=None
    )
    MASTERSHEET_URL: AnyUrl = Field(
        description="The URL pointing to your agency mastersheet.", default=None
    )
    GOOGLE_PROJECT_ID: str = Field(
        description="Google projectID from auth file.", default=None
    )
    GOOGLE_PRIVATE_KEY_ID: str = Field(
        description="Google PrivateKeyId from auth file.", default=None
    )
    GOOGLE_PRIVATE_KEY: str = Field(
        description="Google PrivateKey from auth file.", default=None
    )
    GOOGLE_CLIENT_ID: str = Field(
        description="Google client ID from auth file.", default=None
    )
    GOOGLE_CLIENT_EMAIL: str = Field(
        description="Google client Email from auth file.", default=None
    )
    TIMEZONE: str = Field(
        description="A valid Pytz timezone.", default="America/New_York"
    )
    ALLOWED_SCRAPER_NAMES: list = Field(
        description="A list of scraper names you want to allow.", default=[]
    )
    MAX_WORKERS: int = Field(
        description="The number of concurrent rows you want to process,"
        " this should not be higher than your "
        "concurrency limit from your proxy provider.",
        default=5,
    )
    DATA_CHUNK_SIZE: int = Field(
        default=200,
        description="The number of rows to process "
        "at once in any of the 'process' methods.",
    )
    SCRAPER_MAPPING: dict = Field(
        description="Mapping where the keys are field names "
        "in your data and the values are what you want to swap it with.",
        default={},
    )
    COPYFACTORY_MAPPING: dict = Field(
        description="Mapping where the keys are field names "
        "in your data and the values are what you want to swap it with.",
        default={},
    )
    SENTENCE_WIZARD_OPTIONS: dict = Field(
        description="Mapping where the keys are the scraper names "
        "and the values are a dict of additional scraper options to pass.",
        default={},
    )
    ENABLE_SCRAPER_REUSE: bool = Field(
        description="This setting determines whether a contact can "
        "have multiple personalized sentences from the same scraper.",
        default=False,
    )
    WEBHOOK_URL: AnyUrl = Field(
        description="A webhook endpoint to be notified of events.", default=None
    )

    @validator("TIMEZONE")
    def validate_timezone(cls, value):
        timezones = pytz.all_timezones_set
        if value not in timezones:
            value = "America/New_York"
        return value

    @validator("MASTERSHEET_URL")
    def validate_mastersheet(cls, url):
        if not url:
            return url
        if not isinstance(url, str):
            raise ValueError("Mastersheet is not a valid spreadsheet.")
        parsed_url = urlparse(url)
        if not parsed_url.scheme:
            raise ValueError("Mastersheet is not a valid spreadsheet.")
        if parsed_url.netloc != "docs.google.com":
            raise ValueError("Mastersheet is not a valid spreadsheet.")
        if "spreadsheets" not in parsed_url.path:
            raise ValueError("Mastersheet is not a valid spreadsheet.")
        return url

    @property
    def GOOGLE_CONFIG(self) -> dict:
        """
        Returns a valid Google dict for authentication or empty dict.
        :return:
        """
        if not all(
            (
                self.GOOGLE_PROJECT_ID,
                self.GOOGLE_PRIVATE_KEY_ID,
                self.GOOGLE_PRIVATE_KEY,
                self.GOOGLE_CLIENT_ID,
                self.GOOGLE_CLIENT_EMAIL,
            )
        ):
            return {}
        return {
            "type": "service_account",
            "project_id": self.GOOGLE_PROJECT_ID,
            "private_key_id": self.GOOGLE_PRIVATE_KEY_ID,
            "private_key": self.GOOGLE_PRIVATE_KEY.replace("\\n", "\n"),
            "client_email": self.GOOGLE_CLIENT_EMAIL,
            "client_id": self.GOOGLE_CLIENT_ID,
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": f"https://www.googleapis.com/"
            f"robot/v1/metadata/x509/{self.GOOGLE_CLIENT_EMAIL}",
        }

    def update_settings(self, new_settings: dict = None, env_file_path: str = None):
        """
        Update contact magic config through either a dict or a env file path.
        """
        new_settings = new_settings or {}
        if env_file_path:
            parsed_values = BaseConfig(_env_file=env_file_path).dict()
            new_settings = parsed_values | new_settings
        if not isinstance(new_settings, dict):
            return self
        for k, v in {
            k: v for k, v in new_settings.items() if k in self.__fields__.keys()
        }.items():
            self.__setattr__(k, v)
        *_, validation_error = validate_model(self.__class__, self.__dict__)
        if validation_error:
            raise validation_error
        return self

    def notify_webhook(self, event_name: str, data: dict):
        if self.WEBHOOK_URL:
            requests.post(self.WEBHOOK_URL, json={"event": event_name, "data": data})

    class Config:
        validate_assignment = True
        env_file = ".env"
        env_file_encoding = "utf-8"


SETTINGS = BaseConfig()



from time import sleep

from gspread import Worksheet

try:
    import gspread
    from gspread.exceptions import APIError
    from oauth2client.service_account import ServiceAccountCredentials
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "You do not have the Sheets extension installed."
        " Run `pip install contact-magic[sheets]`"
    )

from contact_magic.conf.settings import SETTINGS


def retry_req(func: callable, *args):
    sleepy_time = 1
    for _ in range(20):
        sleepy_time *= 2
        try:
            return func(*args)
        except APIError as e:
            res = e.response.json()["error"]
            if code := res.get("code"):
                if int(code) in [503, 429]:
                    sleep(min([sleepy_time, 60]))
                    continue
            raise
        except Exception as e:
            print(f"Error in Gsheet with error '{e}'")


def get_spreadsheet_by_url(url) -> gspread.models.Spreadsheet:
    # use creds to create a client to interact with the Google Drive API
    scope = [
        "https://spreadsheets.google.com/feeds",
        "https://www.googleapis.com/auth/drive",
    ]
    creds = ServiceAccountCredentials.from_json_keyfile_dict(
        SETTINGS.GOOGLE_CONFIG, scope
    )
    client = gspread.authorize(creds)
    return retry_req(client.open_by_url, url)


def get_worksheet_from_spreadsheet(
    sheet: gspread.models.Spreadsheet, work_sheet_name: str
) -> gspread.models.Worksheet:
    return retry_req(sheet.worksheet, work_sheet_name)


def bulk_update(sheet, update_range, data=None):
    return retry_req(sheet.update, update_range, data)


def update_cell(sheet, row, col, value):
    return retry_req(sheet.update_cell, row, col, value)


def clear_sheet(sheet):
    return retry_req(sheet.clear)


def format_range(sheet, cell_range: str, format_data: dict):
    return retry_req(sheet.format, cell_range, format_data)


def get_cell_from_sheet(sheet, row, column):
    return retry_req(sheet.cell, row, column)


def get_all_values_from_sheet(worksheet: Worksheet):
    return retry_req(worksheet.get_all_values)



import numpy as np
import pandas as pd

from contact_magic.conf.settings import SETTINGS
from contact_magic.helpers import worksheet_to_dataframe
from contact_magic.integrations.sheets import (
    get_cell_from_sheet,
    get_spreadsheet_by_url,
    get_worksheet_from_spreadsheet,
    update_cell,
)
from contact_magic.scripts.logger import logger
from contact_magic.utils import is_google_workflow_url_valid


def update_rows_approved_and_remaining():
    workflows_sheet = get_worksheet_from_spreadsheet(
        get_spreadsheet_by_url(SETTINGS.MASTERSHEET_URL), "Workflows"
    )
    workflow_values = workflows_sheet.get_all_values()
    df = pd.DataFrame(data=workflow_values[1:], columns=workflow_values[0]).replace(
        "", np.nan
    )
    for i, row in df.iterrows():
        if not all(field in row for field in ["WorkflowUrl", "ClientName"]):
            continue
        if is_google_workflow_url_valid(row["WorkflowUrl"]):
            data = worksheet_to_dataframe(
                get_spreadsheet_by_url(row["WorkflowUrl"]), "WorkingSheet"
            )
            if not all(field in data.columns for field in ["Website", "is_approved"]):
                continue
            logger.info(
                "updating_rows_appproved",
                row_number=i + 2,
                sequence_name=row["WorkflowName"],
                client_name=row["ClientName"],
                status="STARTING",
            )
            total_rows = len(data.dropna(subset=["Website"]))
            total_approved_data = len(data.loc[data["is_approved"] == "TRUE"])
            if total_rows > 0:
                approved_percentage = round(total_approved_data / total_rows, 2)
            else:
                approved_percentage = 0
            base = 5
            for num, metric in enumerate(
                [total_rows, total_approved_data, approved_percentage], 1
            ):
                cell = get_cell_from_sheet(
                    workflows_sheet, row=i + 2, column=base + num
                )
                cell.value = metric
                update_cell(workflows_sheet, cell.row, cell.col, cell.value)
            logger.info(
                "updating_rows_appproved",
                row_number=i + 2,
                sequence_name=row["WorkflowName"],
                client_name=row["ClientName"],
                status="COMPLETE",
            )


if __name__ == "__main__":
    update_rows_approved_and_remaining()



"""
Helpers for working with potentially heavily nested pieces of data
"""


def get_first_level_items(data: dict) -> dict:
    """
    Returns only the top level of data from a dict that is valid.
    :param data:
    :return:
    """
    return {
        key: value for key, value in data.items() if not isinstance(value, (dict, list))
    }


def delete_key_in_object(data, target_key):
    """
    Go through original object and delete
    all occurrences of a target key
    """
    if isinstance(data, list):
        for list_item in data:
            delete_key_in_object(list_item, target_key)
    elif isinstance(data, dict):
        if target_key in data:
            del data[target_key]
        for dict_key, dict_value in data.items():
            delete_key_in_object(dict_value, target_key)
    return data


def replace_values_in_object(data, key, value):
    """
    Go through original object and replace
    all occurrences of a key with new value.
    """
    if isinstance(data, list):
        for list_items in data:
            replace_values_in_object(list_items, key, value)
    elif isinstance(data, dict):
        for dict_key, dict_value in data.items():
            if dict_key == key:
                data[key] = value
            replace_values_in_object(dict_value, key, value)
    return data


def replace_keys_in_object(data, mapping: dict) -> dict:
    """
    Go through the original object and swap any keys to the target mapping.
    """

    result = {}

    for key, value in data.items():
        if isinstance(value, dict):
            result[mapping.get(key, key)] = replace_keys_in_object(value, mapping)
        elif isinstance(value, list):
            result[mapping.get(key, key)] = [
                replace_keys_in_object(item, mapping)
                for item in value
                if isinstance(item, dict)
            ]
        else:
            result[mapping.get(key, key)] = value

    return result


def get_values_in_object_by_key(data, key):
    return list(_get_values_in_object_by_key(data, key))


def _get_values_in_object_by_key(data, key):
    """
    Go through original object and get
    all values based on target key
    """
    if isinstance(data, list):
        for d in data:
            yield from _get_values_in_object_by_key(d, key)
    if isinstance(data, dict):
        for k, v in data.items():
            if key == k:
                yield v
            if isinstance(v, dict):
                yield from _get_values_in_object_by_key(v, key)
            elif isinstance(v, list):
                for d in v:
                    yield from _get_values_in_object_by_key(d, key)



import gspread
import numpy as np
import pandas as pd

from contact_magic.conf.settings import SETTINGS
from contact_magic.integrations.sheets import (
    get_all_values_from_sheet,
    get_worksheet_from_spreadsheet,
)
from contact_magic.models import PersonalizationSettings


def prepare_data_for_gsheet(
    df: pd.DataFrame, mapping: dict = None, enforced_columns: list = None
):
    """
    Converts a dataframe to a list of list to be passed to Gspread or Sheet alternative.
    Can pass a mapping where the key names are columns in the Dataframe.
        The values represent an object of data to be switched.
        Ex: mapping={"is_approved": {'TRUE': True, 'FALSE': False}
    Handy for when converting 'TRUE' as a sting into True as a boolean for
    consistent types in the Google sheet.

    Enforced columns is a list of columns where if rows don't have all the
    columns in the list the row is dropped.
    """
    if isinstance(mapping, dict):
        for col_name, map_conversion in mapping.items():
            if col_name in df:
                df.loc[:, col_name] = df[col_name].map(map_conversion)
    if enforced_columns := enforced_columns or []:
        df = df[df[enforced_columns].notna().all(axis=1)]
    df = df.fillna("")
    return [df.columns.values.tolist()] + df.values.tolist()


def get_personalization_settings_from_sheet(
    main_workflow_sheet: gspread.Spreadsheet, worksheet_name="settings"
) -> PersonalizationSettings:
    """
    Converts a google sheet into a PersonalizationSettings instance.
    """
    settings = []
    settings_data = get_worksheet_from_spreadsheet(
        main_workflow_sheet, worksheet_name
    ).get_all_values()
    df_settings_data = pd.DataFrame(
        data=settings_data[1:], columns=settings_data[0]
    ).replace("", np.nan)

    for i in range(0, len(df_settings_data.columns), 2):
        scraper_col_values = list(df_settings_data.iloc[:, i].values)
        if not scraper_col_values:
            continue
        new_entry = {"col_name": df_settings_data.columns[i], "sentence_wizards": []}
        scrapers_seen = set()
        for position, scraper in enumerate(scraper_col_values):
            if SETTINGS.ALLOWED_SCRAPER_NAMES and (
                scraper not in SETTINGS.ALLOWED_SCRAPER_NAMES
                or scraper in scrapers_seen
            ):
                continue
            premise_url = df_settings_data.iloc[position, i + 1]
            if pd.isnull(premise_url):
                continue
            if scraper == "FALLBACK":
                new_entry["sentence_wizards"].append(
                    {"scraper_name": scraper, "fallback_template": premise_url}
                )
            elif scraper == "USE_HISTORIC_DATA":
                new_entry["sentence_wizards"].append(
                    {
                        "scraper_name": None,
                        "premise_url": premise_url,
                        "restrict_to_scraped_data": False,
                    }
                )
            else:
                new_entry["sentence_wizards"].append(
                    {"scraper_name": scraper, "premise_url": premise_url}
                )
            scrapers_seen.add(scraper)
        if new_entry.get("sentence_wizards"):
            settings.append(new_entry)
    return PersonalizationSettings(datapoints=settings)


def worksheet_to_dataframe(
    main_workflow_sheet: gspread.Spreadsheet, tab_name="WorkingSheet"
) -> pd.DataFrame:
    """
    Convert a tab into a Dataframe.
    """
    worksheet = get_worksheet_from_spreadsheet(main_workflow_sheet, tab_name)
    tab_working_data = get_all_values_from_sheet(worksheet)
    return pd.DataFrame(data=tab_working_data[1:], columns=tab_working_data[0]).replace(
        "", np.nan
    )



import asyncio
import re
from datetime import datetime
from typing import Any
from uuid import UUID, uuid4

import numpy as np
import pandas as pd
from pydantic import AnyUrl, BaseModel, Field, validator

from contact_magic.asyncio import do_bulk
from contact_magic.conf.settings import SETTINGS
from contact_magic.dict_utils import (
    delete_key_in_object,
    get_values_in_object_by_key,
    replace_keys_in_object,
    replace_values_in_object,
)
from contact_magic.integrations import (
    make_copyfactory_request,
    make_sales_scraper_request,
)
from contact_magic.logger import logger
from contact_magic.utils import get_id_from_url, is_valid_premise_url

SALES_SCRAPER_KEY_PREFIX_NAME = "sales_scrapers"


class DataRow(BaseModel):
    """
    A class to hold a row of data and it's index number
    to sort by after. Converts set data to a dict.
    """

    data: Any
    index: int
    _original_type = None

    @validator("data")
    def validate_data(cls, value):
        if not isinstance(value, pd.Series) and not isinstance(value, dict):
            raise ValueError("Data must be either type pd.Series or Dict.")
        _original_type = pd.Series if isinstance(value, pd.Series) else dict
        return value.to_dict() if isinstance(value, pd.Series) else value

    @property
    def convert_to_original(self):
        return (
            pd.Series(data=self.data) if self._original_type == pd.Series else self.data
        )


class SentenceWizard(BaseModel):
    """
    A reference to a scraper endpoint in SalesScrapers and a Copyfactory premiseURL.
    """

    scraper_name: str | None = Field(
        description="This is the endpoint from sales-scrapers", example="get-case-study"
    )
    premise_url: AnyUrl = Field(
        description="The Copyfactory premise URL you want to use. "
        "The ID is extracted from the URL",
        default=None,
    )
    premise_id: int = Field(
        description="The Copyfactory premise ID you want to use.", default=None
    )
    fallback_template: str = ""
    scraper_mapping: dict = Field(
        description="A mapping where the keys are your column "
        "names and the values are the target parameter "
        "name for SalesScrapers.",
        default={},
    )
    copyfactory_mapping: dict = Field(
        description="A mapping where the keys are your column names and the "
        "values are the target variable name you want to send to Copyfactory.",
        default={},
    )
    field_processors: dict = Field(
        description="A list of fields to validate before "
        "doing Copyfactory request."
        "If any of the fields fail validation "
        "Copyfactory will not be called.",
        default={},
    )
    restrict_to_scraped_data: bool = Field(
        description="A boolean to indicate what data the "
        "wizard is allowed to look at before calling Copyfactory."
        " defaults to only looking at the latest scrape attempt. If set to False will "
        "look at all scraped data on the row.",
        default=True,
    )
    added_key_names: set = Field(
        description="A list of keys that were added during the scraping.", default=set()
    )

    @validator("premise_url")
    def validate_premise_url(cls, url):
        if not isinstance(url, str):
            return url
        return url if url.endswith("/") else f"{url}/"

    @validator("scraper_name")
    def validate_scraper_name(cls, value):
        if value is None:
            return value
        if not SETTINGS.ALLOWED_SCRAPER_NAMES:
            return value
        if value not in SETTINGS.ALLOWED_SCRAPER_NAMES:
            raise ValueError(
                f"{value} is not an allowed scraper name "
                f"your configured scrapers are {SETTINGS.ALLOWED_SCRAPER_NAMES}."
            )
        return value

    @property
    def is_premise_url_valid(self) -> bool:
        return is_valid_premise_url(self.premise_url)

    def fill_fallback(self, row: DataRow) -> str:
        """
        Fills in the template with row data.
        either returns the filled in template or an empty string.
        """
        list_pot_keys = re.findall(r"\{(.*?)\}", self.fallback_template)
        copy_of_template = self.fallback_template
        for key in list_pot_keys:
            if value := get_values_in_object_by_key(row.data, key):
                copy_of_template = copy_of_template.replace("{" + key + "}", value[0])
            else:
                return ""
        return copy_of_template

    async def run_sales_scraper(self, row: DataRow, col_prefix: str, overwrite=False):
        """
        Runs the scraping and extends the row by adding
        the column name + scraper as a prefix to ensure uniqueness.
        """
        if not self.scraper_name:
            return row
        if not self.scraper_mapping and SETTINGS.SCRAPER_MAPPING:
            self.scraper_mapping = SETTINGS.SCRAPER_MAPPING

        data = (
            replace_keys_in_object(row.data, self.scraper_mapping)
            if self.scraper_mapping
            else row.data
        )
        if scrape := await make_sales_scraper_request(self.scraper_name, data=data):
            for key, value in scrape.items():
                if key in row.data and not overwrite:
                    continue
                # Remove output schema since don't want
                # to pass schema values to Copyfactory
                # Scope scraper info to a prefixed keyname
                # to not have conflict with other scraped values.
                if key == "scraper_info":
                    if isinstance(value, dict):
                        value = {
                            f"{SALES_SCRAPER_KEY_PREFIX_NAME}__{k}": v
                            for k, v in value.items()
                        }
                value = delete_key_in_object(
                    value, f"{SALES_SCRAPER_KEY_PREFIX_NAME}__output_schema"
                )
                unique_key = f"{col_prefix}__{self.scraper_name}__{key}"
                row.data[unique_key] = value
                self.added_key_names.add(unique_key)
            return row
        return row

    async def run_copyfactory(
        self,
        row: DataRow,
        content_col_name: str,
        source_col_name: str,
        keys_to_delete: set = None,
    ):
        """
        Runs Copyfactory using the current row Data.
        If successfully called with Copyfactory returns True to stop iteration
        for that datapoint.
        """
        keys_to_delete = keys_to_delete or set()
        if not self.premise_url and not self.premise_id:
            return row, False
        if self.premise_url and self.is_premise_url_valid:
            self.premise_id = get_id_from_url(self.premise_url)
        if not self.copyfactory_mapping and SETTINGS.COPYFACTORY_MAPPING:
            self.copyfactory_mapping = SETTINGS.COPYFACTORY_MAPPING

        row, is_valid = self.run_processors(row)
        if not is_valid:
            return row, False

        data = (
            replace_keys_in_object(row.data, self.copyfactory_mapping)
            if self.copyfactory_mapping
            else row.data
        )
        # Remove all other scraped data keys
        # from the row from consideration in the event of
        # multiple keys from scraped data having
        # the same name to restrict the passed data to being from
        # this sentence wizard scrape.
        if self.restrict_to_scraped_data:
            for key in keys_to_delete:
                if not key.startswith(f"{content_col_name}__{self.scraper_name}"):
                    data = delete_key_in_object(data, key)

        if cf := await make_copyfactory_request(self.premise_id, variables=data):
            row.data[content_col_name] = cf["content"]
            source = row.data.get(
                f"{content_col_name}__{self.scraper_name}__scraper_info", {}
            ).get(f"{SALES_SCRAPER_KEY_PREFIX_NAME}__data_source", "-")
            row.data[source_col_name] = f"{self.scraper_name}, {source}"
            return row, True
        return row, False

    def run_processors(self, row) -> tuple[DataRow, bool]:
        """
        Iterate over field processors and break if any don't satisfy.
        """

        if not self.field_processors:
            return row, True

        from contact_magic.preprocessors import processors

        for key, processor in self.field_processors.items():
            if item := get_values_in_object_by_key(row.data, key):
                if validation_model := processors.get(
                    processor.get("type", None), None
                ):
                    # If Pydantic fails validation/init just move on.
                    try:
                        value, is_value_valid = validation_model(
                            processor=processor,
                            value=item[0],
                            wizard=self,
                            row_data=row.data,
                        ).process()
                        row.data = replace_values_in_object(row.data, key, value)
                    except Exception:
                        is_value_valid = False
                    if not is_value_valid:
                        return row, False
        return row, True

    async def execute(
        self,
        row: DataRow,
        content_col_name: str,
        source_col_name: str,
        keys_to_delete: set = None,
    ) -> tuple[DataRow, bool]:
        """
        Executes the scraper and extends the DataRow as it moves along to
        allow for exploding the DF with more
        datapoints from the scraping + personalization.
        """
        keys_to_delete = keys_to_delete or set()
        if self.scraper_name == "FALLBACK" and self.fallback_template:
            row.data[content_col_name] = self.fill_fallback(row)
            row.data[source_col_name] = "-"
            return row, row.data[content_col_name] != ""
        row = await self.run_sales_scraper(row, col_prefix=content_col_name)
        row, success = await self.run_copyfactory(
            row, content_col_name, source_col_name, keys_to_delete=keys_to_delete
        )
        return row, success

    def __init__(self, **data: Any):
        if data.get("scraper_name", None):
            if default_options := SETTINGS.SENTENCE_WIZARD_OPTIONS.get(
                data.get("scraper_name"), None
            ):
                data |= default_options
        super().__init__(**data)


class DataPoint(BaseModel):
    """
    A reference to a datapoint to create for a given contact
    targeting a column and a list of allowed scrapers.
    """

    col_name: str = Field(
        description="The column name you want for the datapoint.",
        example="Personalization1",
    )
    sentence_wizards: list[SentenceWizard]
    keys_added: set = set()
    successful_scraper_name: str = ""

    @property
    def has_successful_sentence(self):
        return len(self.successful_scraper_name) > 0

    @property
    def get_col_name_as_source(self) -> str:
        return f"source__{self.col_name}"

    async def build_datapoint(
        self, row: DataRow, keys_to_delete: set = None, wizards_to_remove: set = None
    ):
        keys_to_delete = keys_to_delete or set()
        wizards_to_remove = wizards_to_remove or set()
        if wizards_to_remove:
            wizards = [
                wizard
                for wizard in self.sentence_wizards
                if wizard.scraper_name not in wizards_to_remove
            ]
        else:
            wizards = self.sentence_wizards
        # remove any wizards for this row that have been used.
        for scraper in wizards:
            logger.info(
                "processing_row",
                row_number=row.index,
                scraper_name=scraper.scraper_name,
                premise_id=scraper.premise_id or get_id_from_url(scraper.premise_url),
                column_name=self.col_name,
                status="STARTING",
            )
            row, did_succeed = await scraper.execute(
                row,
                self.col_name,
                self.get_col_name_as_source,
                keys_to_delete=keys_to_delete | self.keys_added,
            )
            for key in scraper.added_key_names:
                self.keys_added.add(key)
            if did_succeed:
                if scraper.scraper_name != "FALLBACK":
                    self.successful_scraper_name = scraper.scraper_name
                logger.info(
                    "processing_row",
                    row_number=row.index,
                    scraper_name=scraper.scraper_name,
                    premise_id=scraper.premise_id
                    or get_id_from_url(scraper.premise_url),
                    column_name=self.col_name,
                    status="SUCCESS",
                )
                break
            else:
                logger.info(
                    "processing_row",
                    row_number=row.index,
                    scraper_name=scraper.scraper_name,
                    premise_id=scraper.premise_id
                    or get_id_from_url(scraper.premise_url),
                    column_name=self.col_name,
                    status="FAILED",
                )
        return row


class PersonalizationSettings(BaseModel):
    """
    A reference to a list of datapoints to generate for a given contact.
    """

    name: str = ""
    datapoints: list[DataPoint]
    uid: UUID = Field(default_factory=uuid4)
    created: datetime = Field(default_factory=datetime.utcnow)

    @validator("datapoints")
    def validate_datapoints(cls, datapoints):
        seen_names = {}
        for dp in datapoints:
            if dp.col_name in seen_names:
                seen_names[dp.col_name] += 1
                dp.col_name = f"{dp.col_name}__{seen_names[dp.col_name]}"
            else:
                seen_names[dp.col_name] = 1
        return datapoints

    @property
    def get_datapoint_column_names(self) -> list:
        """
        Returns a list of column names and a source field.
        """
        return [dp.col_name for dp in self.datapoints] + [
            dp.get_col_name_as_source for dp in self.datapoints
        ]

    def extend_dataframe_with_settings_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(
            **{
                c: np.nan
                for c in self.get_datapoint_column_names
                if c not in df.columns
            }
        )

    def process_from_dataframe(
        self, df: pd.DataFrame, exclude_filter_func: callable = None
    ):
        """
        Process a dataframe and extend it's rows based on the
        personalization settings.
        Also accepts a callable which takes as its only argument
        the row from the dataframe to filter on.
        This function must resolve to a boolean and rows that are
        True are not processed.
        """
        df = self.extend_dataframe_with_settings_columns(df)
        untouched_rows = []
        jobs = []
        # Set to 1 based index for logging
        df.index = df.index + 1
        for i, row in df.iterrows():
            if exclude_filter_func and exclude_filter_func(row):
                untouched_rows.append(DataRow(data=row, index=i))
                continue
            jobs.append(DataRow(data=row, index=i))
        return pd.DataFrame(
            data=[
                row.convert_to_original
                for row in sorted(
                    untouched_rows + self.process_data_rows(jobs),
                    key=lambda dp: dp.index,
                )
            ],
            columns=df.columns,
        )

    def process_data_rows(self, data: list[DataRow]) -> list[DataRow]:
        """
        Process DataRows async.
        :param data:
        :return:
        """
        return asyncio.run(
            do_bulk(
                [(self.build_row, {"row": row}) for row in data],
                max_workers=SETTINGS.MAX_WORKERS,
            )
        )

    async def build_row(self, row: DataRow):
        """
        Iterate over all personalization datapoints and their
        allowed scrapers to extend a row of contact data
        with enrichment and personalized sentences.
        """
        keys_to_delete = set()
        wizards_to_remove = set()

        for datapoint in self.datapoints:
            row = await datapoint.build_datapoint(
                row, keys_to_delete=keys_to_delete, wizards_to_remove=wizards_to_remove
            )
            if SETTINGS.ENABLE_SCRAPER_REUSE is False:
                if datapoint.has_successful_sentence:
                    wizards_to_remove.add(datapoint.successful_scraper_name)
            for key in datapoint.keys_added:
                keys_to_delete.add(key)
        return row



import asyncio
from asyncio import QueueEmpty

from contact_magic.conf.settings import SETTINGS
from contact_magic.logger import logger


async def do_bulk(ops: list, max_workers: int = SETTINGS.MAX_WORKERS):
    """
    Bulk operation
    This function can be used to run bulk operations
    using a limited number of concurrent requests.
    """
    results = [None for _ in range(len(ops))]

    queue = asyncio.Queue()

    for job in enumerate(ops):
        await queue.put(job)

    workers = [_worker(queue, results) for _ in range(max_workers)]

    await asyncio.gather(*workers)

    return results


async def _worker(queue: asyncio.Queue, results: list):
    while True:
        try:
            index, op = queue.get_nowait()
        except QueueEmpty:
            break

        try:
            response = await op[0](**op[1])
            results[index] = response
        except Exception as e:
            logger.error(
                "processing_row",
                row_number=op[1]["row"].index,
                status="ERROR",
                message=e,
            )
            results[index] = op[1]["row"]
        queue.task_done()



import datetime

try:
    import gspread
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "You do not have the Sheets extension installed."
        " Run `pip install contact-magic[sheets]`"
    )
import numpy as np
import pandas as pd
import pytz

from contact_magic.conf import settings as global_settings
from contact_magic.conf.settings import SETTINGS
from contact_magic.helpers import (
    get_personalization_settings_from_sheet,
    prepare_data_for_gsheet,
    worksheet_to_dataframe,
)
from contact_magic.integrations.sheets import (
    bulk_update,
    format_range,
    get_spreadsheet_by_url,
    get_worksheet_from_spreadsheet,
    update_cell,
)
from contact_magic.scripts.logger import logger
from contact_magic.utils import chunk_df, convert_to_a1, is_google_workflow_url_valid


def get_workflows_to_run(sheet):
    """
    Filter for only active workflows & ones with workflows URLs.
    """
    workflow_values = sheet.get_all_values()
    df = pd.DataFrame(data=workflow_values[1:], columns=workflow_values[0]).replace(
        "", np.nan
    )
    workflows_to_run = df.loc[df["RunWorkflow"] == "TRUE"]
    return workflows_to_run[workflows_to_run["WorkflowUrl"].notna()]


def update_date_last_ran(
    worksheet: gspread.Worksheet, row_number: int, col_number: int = 5
):
    """
    Update a cell with the latest date based on the configured timezone.
    """
    timezone = pytz.timezone(SETTINGS.TIMEZONE)
    current_time = datetime.datetime.now(timezone).strftime("%Y-%m-%d %H:%M:%S")
    cell = worksheet.cell(row=row_number, col=col_number)
    cell.value = current_time
    update_cell(worksheet, cell.row, cell.col, cell.value)


def filter_out_row(row) -> bool:
    if row["is_approved"] == "TRUE":
        return True
    if pd.isnull(row["Website"]):
        return True
    return False


def uncheck_rows_and_format(sheet, df: pd.DataFrame, col_number=4):
    for i, row in df.iterrows():
        row_num = i + 2
        update_cell(sheet, row_num, col_number, False)
        format_range(
            sheet,
            f"{row_num}:{row_num}",
            {
                "backgroundColor": {"red": 1.0, "green": 1.0, "blue": 0.3},
            },
        )


def mark_row_as_completed(sheet, row_number):
    row_num = row_number + 2
    format_range(
        sheet,
        f"{row_num}:{row_num}",
        {
            "backgroundColor": {"red": 0.0, "green": 1.0, "blue": 0.0},
        },
    )


def run_sheets(return_as_arguments=False):
    workflows_sheet = get_worksheet_from_spreadsheet(
        get_spreadsheet_by_url(SETTINGS.MASTERSHEET_URL), "Workflows"
    )
    workflows_to_run = get_workflows_to_run(workflows_sheet)
    uncheck_rows_and_format(workflows_sheet, workflows_to_run)

    tasks = []
    for i, row in workflows_to_run.iterrows():
        if not is_google_workflow_url_valid(row["WorkflowUrl"]):
            continue
        workflow_sheet = get_spreadsheet_by_url(row["WorkflowUrl"])
        filtered_working_data = worksheet_to_dataframe(workflow_sheet)
        # Don't filter working data since need to maintain
        # index so do spoof check to know if any rows to process.
        if (
            filtered_working_data.loc[filtered_working_data["is_approved"] == "FALSE"]
            .dropna(subset=["Website"])
            .empty
        ):
            continue
        if return_as_arguments:
            tasks.append(
                (
                    process_worksheet,
                    {
                        "filtered_working_data": filtered_working_data,
                        "i": i,
                        "row": row,
                        "workflow_sheet": workflow_sheet,
                        "workflows_sheet": workflows_sheet,
                    },
                )
            )
        else:
            process_worksheet(
                filtered_working_data, i, row, workflow_sheet, workflows_sheet
            )
    return tasks


def process_worksheet(filtered_working_data, i, row, workflow_sheet, workflows_sheet):
    logger.info(
        "running_workflow",
        row_number=i + 2,
        dataset_size=len(filtered_working_data),
        sequence_name=row["WorkflowName"],
        client_name=row["ClientName"],
        status="STARTING",
    )
    if all(col in filtered_working_data.columns for col in ["Company Name", "City"]):
        filtered_working_data["search_query"] = filtered_working_data[
            "Company Name"
        ].str.cat(filtered_working_data["City"], sep=" ")
    if all(
        col in filtered_working_data.columns for col in ["City", "State", "Country"]
    ):
        filtered_working_data["location_search_from"] = filtered_working_data[
            "City"
        ].str.cat(
            [filtered_working_data["State"], filtered_working_data["Country"]],
            sep=", ",
        )
    campaign_settings = get_personalization_settings_from_sheet(workflow_sheet)
    offset = 0
    working_sheet = get_worksheet_from_spreadsheet(workflow_sheet, "WorkingSheet")
    for chunk in chunk_df(
        filtered_working_data,
        global_settings.DATA_CHUNK_SIZE or len(filtered_working_data),
    ):
        processed_data = campaign_settings.process_from_dataframe(
            df=chunk, exclude_filter_func=filter_out_row
        )
        data = prepare_data_for_gsheet(
            processed_data,
            {"is_approved": {"TRUE": True, "FALSE": False}},
            enforced_columns=["Website"],
        )
        if offset == 0:
            bulk_update(
                working_sheet,
                f"A1:{convert_to_a1(row=1, col=len(processed_data.columns))}",
                [data[0]],
            )
        row_data = data[1:]
        start = convert_to_a1(row=2 + offset, col=1)
        offset += global_settings.DATA_CHUNK_SIZE
        if not row_data:
            continue
        row_num = len(row_data) + 1 + offset
        if row_num > len(filtered_working_data):
            row_num = len(filtered_working_data) + 1
        end = convert_to_a1(row=row_num, col=len(processed_data.columns))
        bulk_update(working_sheet, f"{start}:{end}", row_data)

    update_date_last_ran(workflows_sheet, i + 2)
    logger.info(
        "running_workflow",
        row_number=i + 2,
        dataset_size=len(filtered_working_data),
        sequence_name=row["WorkflowName"],
        client_name=row["ClientName"],
        status="COMPLETE",
    )
    row = row.fillna("")
    SETTINGS.notify_webhook(
        "worksheet_complete",
        {
            "client_name": row["ClientName"],
            "sequence_name": row["WorkflowName"],
            "workflow_url": row["WorkflowUrl"],
        },
    )
    mark_row_as_completed(workflows_sheet, i)


if __name__ == "__main__":
    run_sheets()



import pandas as pd

fallback_description = """
Replaces a text pattern Ex:'P.S {first_name}
what did you think?' with data from the contact. In this example, {first_name}
would be swapped with the column value of 'first_name' in your contact data.
"""

historic_data_description = """
Doesn't perform any scrape and uses
the current contact data and any scraped data
to generate a sentence using Copyfactory.
"""

case_study_description = """
This scraper aims to gather data about a
case study that the target company has worked on.
The data points you can expect to retrive are the following:
Case Study Overview, One Sentence Case Study Summary,
The company mentioned in the case study
and the source of where the data was found.
"""

company_achievement_description = """
This scraper aims to gather data
about a recent company achievement.
The data points you can expect to
retrieve are the following: Company Achievement Overview,
One Sentence Achievement Summary,
The company mentioned in the Achievement and
the source of where the data was found.
"""

company_description_description = """
This scraper aims to gather data about the comapnies description of what they do.
The data points you can expect to retrieve are the following:
Company Description, One Sentence Description Summary,
The company mentioned in the Description and the source of where the data was found.
"""

company_announcement_description = """
This scraper aims to gather data about a recent company announcement.
The data points you can expect to retrieve are the following:
 Company Announcement, One Sentence Announcement Summary,
 The company mentioned in the Announcement and the source of where the data was found.
"""

search_google_description = """
This scraper aims to gather data about a Google Search.
The data points you can expect to retrieve are the following:
 Customer Review, SEO Ranking of website for search term,
 The company mentioned in the result, The company description,
 The Title of the page and the search term used for the search
 and the source of where the data was found.
"""
get_linked_profile_description = """
This scraper aims to gather data about a prospect from their LinkedIn profile.
The data points you can expect to retrieve are the following: FIrst Name,
 Last Name, Company Name, City, State, Country, Current Employment Name,
  Current Title, Start Date, Company Description, Previous employment
  company name, Previous employment title, Previous employment start date,
  Previous employment end date, Previous employment company description,
  Education Establishment Name, Education Degree Name, Education start date,
  Education End Date,  Languages they speak, time in current role, did they
  recently change job, did they recently get promoted, LinkedIn profile URL.
"""
company_profile_description = """
Gathers data about the company public Linkedin profile.
"""
default_scraper_options = pd.DataFrame(
    data=[
        {
            "Scraper": "FALLBACK",
            "Scraper Description": fallback_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "USE_HISTORIC_DATA",
            "Scraper Description": historic_data_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "get-case-study",
            "Scraper Description": case_study_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "get-company-achievement",
            "Scraper Description": company_achievement_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "get-company-description",
            "Scraper Description": company_description_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "get-company-announcement",
            "Scraper Description": company_announcement_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "search-google",
            "Scraper Description": search_google_description.strip(),
            "Notes": "",
        },
        {
            "Scraper": "get-person-profile",
            "Scraper Description": get_linked_profile_description.strip(),
            "Notes": "This can be expensive to use please "
            "check estimate for scraping LinkedIn",
        },
        {
            "Scraper": "get-company-profile",
            "Scraper Description": company_profile_description.strip(),
            "Notes": "",
        },
    ]
)



"""
AGENCY CLI for running and interacting with the scripts.
"""
from enum import Enum
from typing import List

import typer
from rich import print
from typing_extensions import Annotated

from contact_magic.scripts.run_workflows import run_sheets
from contact_magic.scripts.sync_scraper_options_to_workersheets import (
    sync_scraper_options,
)
from contact_magic.scripts.update_workflow_approval_metrics import (
    update_rows_approved_and_remaining,
)

app = typer.Typer()


class ScriptOptions(Enum):
    run_sheets = "run_sheets"
    update_approved = "update_approved"
    sync_scraper_options = "sync_scraper_options"


@app.command()
def pipeline(
    task: List[ScriptOptions] = None,
    run_all: Annotated[bool, typer.Option("--run-all")] = False,
    times_to_run: Annotated[int, typer.Option(min=1)] = 1,
):
    """
    Run a list of scripts in a single command.

    Notice how you can set the order and number of times to run.

    Ex: "python agency.py --task run_sheets --task run_sheets --task update_approved"

    To run all of the scripts

    Ex: "python agency.py --run-all"
    """
    commands = {
        "run_sheets": run_sheets,
        "update_approved": update_rows_approved_and_remaining,
        "sync_scraper_options": sync_scraper_options,
    }
    tasks = task or []
    for _ in range(times_to_run):
        if run_all:
            for name, command in commands.items():
                print(f":robot_face: [green]Running[/green] {name}")
                command()
                print(f":robot_face: [green]Completed[/green] {name}")
            continue
        for task in tasks:
            print(f":robot_face: [green]Running[/green] {task.value}")
            commands[task.value]()
            print(f":robot_face: [green]Completed[/green] {task.value}")



import asyncio
import logging

import httpx

from contact_magic.conf.settings import SETTINGS
from contact_magic.dict_utils import get_first_level_items, get_values_in_object_by_key
from contact_magic.logger import logger

logging.getLogger("httpx").setLevel(logging.WARNING)


endpoint = "https://app.copyfactory.io/api/v2/generate/"


def _get_rate_limit_sleep_time(response):
    """Get rate limit window expiration time from response if the response
    status code is 429.
    """
    try:
        data = response.headers
        if "Retry-After" in data.keys():
            return int(data["Retry-After"])
    except (AttributeError, KeyError, ValueError):
        return 60


async def make_copyfactory_request(premise_id: int, variables: dict, max_retries=3):
    if not SETTINGS.COPYFACTORY_API_KEY:
        return None
    headers = {
        "Accept": "application/json",
        "Authorization": SETTINGS.COPYFACTORY_API_KEY,
    }
    variable_data = get_first_level_items(variables)
    data = {
        "premise_id": premise_id,
        "variables": variable_data,
    }
    try:
        session = httpx.AsyncClient(timeout=90)
        retries = 0
        while retries < max_retries:
            res = await session.request(
                method="POST", url=endpoint, headers=headers, json=data
            )
            # authorization error so can break
            if res.status_code == 403:
                logger.warning("copyfactory_error", message=res.text)
                break
            if res.status_code == 429:
                await asyncio.sleep(_get_rate_limit_sleep_time(res))
                retries += 1
                continue

            if res.status_code == 200:
                response_data = res.json()["data"]
                if response_data["status"] == "success":
                    return response_data
                if response_data["status"] == "error":
                    required_variables = response_data["meta_data"]["sentence_premise"][
                        "required_variables"
                    ]
                    for req_var in required_variables:
                        if item := list(
                            get_values_in_object_by_key(variables, req_var)
                        ):
                            data["variables"][req_var] = str(item[0])
                    if not data["variables"].get("company_organization_name"):
                        data["variables"]["company_organization_name"] = ""

            retries += 1
        return None
    except Exception:
        return None



import importlib.metadata

__version__ = importlib.metadata.version("contact-magic")

from .conf.settings import SETTINGS as settings  # noqa
from .models import DataPoint, DataRow, PersonalizationSettings, SentenceWizard  # noqa
from .scripts.agency import app as agency  # noqa



import re
from abc import abstractmethod
from typing import Any

from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.main import validate_model

from contact_magic import SentenceWizard


class BaseProcessor(BaseModel):
    """
    Base validator to inherit from. Define the 'validate' method to return a boolean.
    """

    value: Any = None
    processor: dict
    wizard: SentenceWizard
    row_data: dict

    _is_valid: bool = True

    def process(self) -> tuple[Any, bool]:
        if not self._is_valid:
            return self.value, self._is_valid
        return self.validate(**self.processor)

    @abstractmethod
    def validate(self, *args, **kwargs) -> tuple[Any, bool]:
        return self.value, self._is_valid

    class Config:
        arbitrary_types_allowed = True

        @classmethod
        def prepare_field(cls, field: ModelField) -> None:
            # Ensure any self declared validation models don't require a value.
            if field.name == "value":
                field.required = False

    def __init__(self, **data: Any) -> None:
        """
        Don't raise validation error if can't coerce,
        just set value to None which fails validation.
        """
        values, fields_set, validation_error = validate_model(self.__class__, data)
        if validation_error:
            self._is_valid = False
        super().__init__(**data)


class StringProcessor(BaseProcessor):
    value: str

    def validate(
        self,
        min_length=None,
        max_length=None,
        regex=None,
        choices=None,
        *args,
        **kwargs
    ) -> tuple[Any, bool]:
        if min_length is not None and len(self.value) < min_length:
            return self.value, False

        if max_length is not None and len(self.value) > max_length:
            return self.value, False

        if regex is not None and not bool(re.match(regex, self.value)):
            return self.value, False

        if choices is not None and self.value not in choices:
            return self.value, False
        return self.value, True


class IntegerProcessor(BaseProcessor):
    value: int

    def validate(
        self, min_amount=None, max_amount=None, *args, **kwargs
    ) -> tuple[Any, bool]:
        if min_amount is not None and self.value < min_amount:
            return self.value, False

        if max_amount is not None and self.value > max_amount:
            return self.value, False

        return self.value, True


class FloatProcessor(BaseProcessor):
    value: float

    def validate(
        self, min_amount=None, max_amount=None, *args, **kwargs
    ) -> tuple[Any, bool]:
        if min_amount is not None and self.value < min_amount:
            return self.value, False

        if max_amount is not None and self.value > max_amount:
            return self.value, False
        return self.value, True


processors = {
    "str": StringProcessor,
    "string": StringProcessor,
    "int": IntegerProcessor,
    "integer": IntegerProcessor,
    "float": FloatProcessor,
}



import asyncio
import logging

import httpx

from contact_magic.conf.settings import SETTINGS
from contact_magic.dict_utils import get_first_level_items
from contact_magic.logger import logger
from contact_magic.utils import fix_website

logging.getLogger("httpx").setLevel(logging.WARNING)

base_url = "https://salesscrapers.dev/api/"


async def make_sales_scraper_request(endpoint, data: dict, max_retries=3):
    if not SETTINGS.SALES_SCRAPERS_API_KEY:
        return None
    headers = {
        "Accept": "application/json",
        "X-API-Key": SETTINGS.SALES_SCRAPERS_API_KEY,
    }

    url = f"{base_url}{endpoint}"
    try:
        session = httpx.AsyncClient(timeout=120)
        retries = 0
        params = get_first_level_items(data)
        while retries < max_retries:
            res = await session.request(
                method="get", url=url, headers=headers, params=params
            )
            # authorization error so can break
            if res.status_code == 401:
                logger.warning("sales_scraper_error", message=res.text)
                break
            if res.status_code == 400:
                logger.warning("proxy_error", message=res.text)
            if res.status_code == 200:
                return res.json()
            if res.status_code == 422:
                validation_errors = res.json().get("detail")
                for error in validation_errors:
                    if error.get("type") == "value_error.url.scheme":
                        error_key = error.get("loc")[1]
                        if error_key in data:
                            params[error_key] = fix_website(data[error_key])
            await asyncio.sleep(1)
            retries += 1
        return None
    except Exception:
        return None



import asyncio

import numpy as np
import pandas as pd

from contact_magic.conf.settings import SETTINGS
from contact_magic.integrations import make_sales_scraper_request
from contact_magic.integrations.sheets import (
    bulk_update,
    clear_sheet,
    get_spreadsheet_by_url,
    get_worksheet_from_spreadsheet,
)
from contact_magic.scripts.default_scraper_options import default_scraper_options
from contact_magic.scripts.logger import logger
from contact_magic.utils import is_google_workflow_url_valid


async def get_user_scrapers_as_df():
    res = await make_sales_scraper_request(endpoint="custom-scrapers", data={})
    user_scrapers_as_df = pd.json_normalize(data=res)
    user_scrapers_as_df = user_scrapers_as_df.rename(
        columns={
            "scraper_slug_name": "Scraper",
            "scraper_properties.description": "Scraper Description",
        }
    )
    user_scrapers_as_df["Scraper"] = (
        user_scrapers_as_df["api_namespace"] + "/" + user_scrapers_as_df["Scraper"]
    )
    user_scrapers_as_df["Notes"] = ""
    return user_scrapers_as_df[["Scraper", "Scraper Description", "Notes"]]


def sync_scraper_options(scraper_options_df=default_scraper_options):
    workflows_sheet = get_worksheet_from_spreadsheet(
        get_spreadsheet_by_url(SETTINGS.MASTERSHEET_URL), "Workflows"
    )
    workflow_values = workflows_sheet.get_all_values()
    df = pd.DataFrame(data=workflow_values[1:], columns=workflow_values[0]).replace(
        "", np.nan
    )
    user_scrapers = asyncio.run(get_user_scrapers_as_df())
    final_df = pd.concat([scraper_options_df, user_scrapers])
    for i, row in df.iterrows():
        if is_google_workflow_url_valid(row["WorkflowUrl"]):
            logger.info(
                "updating_scraper_options",
                row_number=i + 2,
                sequence_name=row["WorkflowName"],
                client_name=row["ClientName"],
                status="STARTING",
            )
            ss = get_spreadsheet_by_url(row["WorkflowUrl"])
            ws = get_worksheet_from_spreadsheet(ss, "scraper_options")
            clear_sheet(ws)
            bulk_update(
                ws,
                [final_df.columns.values.tolist()] + final_df.values.tolist(),
            )
            logger.info(
                "updating_scraper_options",
                row_number=i + 2,
                sequence_name=row["WorkflowName"],
                client_name=row["ClientName"],
                status="COMPLETE",
            )
